# These are general revocation subroutines

package revokelib;

require '/u/mualice/revoke/des-lib.perl';

# Since there is no path (usually) when script is run from slocal, fully
# specify executables.
$md5_path="/u/mualice/revoke/md5"; # Path to MD5 executable
$send_path="/usr/local/pkg/mh/bin/send"; # Path to MH's send
$mhparam_path="/usr/local/pkg/mh/bin/mhparam"; 
$rm_path="/usr/bin/rm";

# Locking constants (read flock man page)
$LOCK_SH = 1;               
$LOCK_EX = 2;
$LOCK_NB = 4;
$LOCK_UN = 8;

sub des # ($unp_hinput, $unp_hkey, $encrypt, $unp_houtput)
# Requires: Exactly 128-bit input data (hex), a 64-bit DES key (hex), and
#           encrypt/decrypt selector.  The DES key should be a string (of
#           length 16) in hex, eg "0123456789abcdef".
# Modifies: $unp_houtput (hex)
# Effects: If $encrypt==1, returns a DES CBC encrypted 64-bit ciphertext
#          where the key works double as the IV.  Otherwise, decrypt.
#          (In the future, may use $IV=0 since only 1 block)
{
    local ($unp_hinput, $unp_hkey, $encrypt, $unp_houtput)=@_;
    local ($schedule,$input,$key,$output)=();
    open (LOGFILE,">$ENV{'HOME'}/desfile");
    print LOGFILE "unp_hinput: >$unp_hinput<\n";
    print LOGFILE "unp_hkey: >$unp_hkey<\n";
    print LOGFILE "encrypt: >$encrypt<\n";
    print LOGFILE "unp_houtput before: >$unp_houtput<\n";
    $input=pack("H*",$unp_hinput);
    $key=pack("H*",$unp_hkey);
    &main'make_key_sched($key,$schedule);
    local($footemp)=unpack("H*",$schedule);
    print LOGFILE "schedule: >$footemp<\n";
    &main'des_cbc_encrypt($input,$output,16,$schedule,$key,$encrypt);
    $_[3]=substr(unpack("H*",$output),0,32);
    print LOGFILE "unp_houtput after: >$_[3]<\n";
    close(LOGFILE);
}

sub md5_message # ($body,$hash,%header)
# Requires: message body (with proper escaping where necessary), header 
#           associative array (where the keys' first letter and all letters
#           preceeding a dash are captialized).
# Modifies: $hash
# Effects: Returns the 128-bit MD5 hash of the message body and certain
#          headers as defined in the (yet to be made) documentation.
{
    local($body,$hash,%header)=@_;

    # Concatenate To (user part), X-Revoke-From, X-Revoke-Date, Subject, body.
    $to=$header{'To'};
#    print "To before: $to\n";
    # The domain isn't always specified until after being processed by a MTA.
    # Plus, case can change.  Just ignore domains and use usernames.
    $to=~s/(.*)\@.*$/$1/;
#    print "To after: $to\n";
    $hashme=$to.$header{'X-Revoke-From'}.$header{'X-Revoke-Date'}.
	$header{'Subject'}.$body;
    open (FOO,">md5log");
    print FOO "hashme: >$hashme<\n"; 
    chop($hash=`echo "$hashme" | $md5_path`);
# I'm not 100% sure about the above quotation.  The shell doesn't deal
# otherwise with CRLF's in the body.
    $_[1]=$hash;    
}

sub parse_message  #(*FILE,$body,$separator,$header_string)
# Requires: an opened filehandle containing a draft or message following RFC822
# Modifies: $body, $separator, $header_string
# Effects: Returns an associative array of all the headers, body, the draft's
#          header-body separator string, a string of names of headers 
#          separated by ":".  A separator is not necessary in the message if
#          no body.
{
    # Assumes message conforms to RFC822 and MH conventions
    # A blank line or one or more beginning line dashes ends header section.
    local(*FILE,$body,$separator,$header_string)=@_;
    local($header_done,$first_body_line,@ordered_header);
    $body="";
    $separator="";
    while ($line=<FILE>) 
    {
	if ((($line=~/^-{1,}/) || ($line=~/^$/)) && (!$header_done))
	    # 6 (or more) dashes or line of whitespace
	{ chop($separator=$line); $header_done=1; $first_body_line=1; }

	if ($header_done) 
	{
	    # Get the body and remove leading whitespace
	    if ($first_body_line)
	    { $first_body_line=0; }
	    else 
	    {$body= $body.$line; }
	} else {
            # Header field names may contain any ASCII chacter from 
            # 33->126 (decimal) except colon
	    # Note that this does not catch all headers...if the header
	    # is multiline, or stuff like To: Name <fubob@foo.com> won't work
	    if ($line=~/^([\x21-\x7E]*):\s*(.*)/i)
		{ 
		    $header{$1}=$2; 
		    push(@ordered_header,$1);
		}
	}
    }
    $_[1]=$body;
    $_[2]=$separator;
    $_[3]=join(":",@ordered_header);
    %header;
}

sub gen_session_key # ($sequence,$unp_hseed)
# Requires: The sequence number and unp_hseed (hex).
# Modifies: $sequence, $unp_hseed
# Effects: Returns a random DES key in hex by encrypting the current 
#          sequence number with the random seed, and then saving the 
#          new value as the next random seed and incrementing the 
#          sequence number.
{
    local($sequence,$unp_hseed)=@_;

    $seed=pack("H*",$unp_hseed);
    # encrypt sequence with seed.
    &main'make_key_sched($seed,$schedule);
    &main'des_cbc_encrypt($sequence,$newseedkey,8,$schedule,$seed,1);
 
    # Store new seed and key
    $_[0]=$sequence+1;
    $_[1]=unpack("H*",$newseedkey);
}

sub get_params # ($sequence,$unp_hseed)
# Modifies: $sequence, $unp_hseed
# Effects: Gets the sequence and seed array from the parameter file
{
    local($sequence,$unp_hseed)=@_;
    chop($mhrevoke=`$mhparam_path mhrevoke`);
    if ($mhrevoke=~/^$/) { $mhrevoke= ".revoke"; }
    chop($path=`$mhparam_path path`);
    if ($path=~/^$/) { $path= "Mail"; }
    open(FILE,"$ENV{'HOME'}/$path/$mhrevoke");
    chop($sequence=<FILE>);
    chop($unp_hseed=<FILE>);
    close(FILE);
    $_[0]=$sequence;
    $_[1]=$unp_hseed;
}

sub store_params # ($sequence,$unp_hseed)
# Effects: Store the sequence and seed from the parameter file
{
    local($sequence,$unp_hseed)=@_;
    chop($mhrevoke=`$mhparam_path mhrevoke`);
    if ($mhrevoke=~/^$/) { $mhrevoke= ".revoke"; }
    chop($path=`$mhparam_path path`);
    if ($path=~/^$/) { $path= "Mail"; }
    open(FILE,">$ENV{'HOME'}/$path/$mhrevoke");
    print FILE "$sequence\n";
    print FILE "$unp_hseed\n";
    close(FILE);
}

sub grab_message # ($spool,$hash,$spool_body,$found);
# Effects: Grabs the body and headers of the unique message in the
#          spool file indexed by the hash.  Returns assoc. array of headers
#          Found is 1 if found. Otherwise not 1.
{
    local($spool,$hash,$spool_body,$found)=@_;
    local(%spool_header,$header_done,$first_body_line);
    $spool_body="";
    # Look for hash match in spool file
    open (SPOOL,$spool);
    while($line=<SPOOL>)
    {   
	if ($line=~/^From\x20/) # beginning of new message
	{   
            # Check if last read message is the "chosen one".
	    if ($hash eq $spool_header{'X-Revoke-Hash'}) # Found it!
	    { 
		$spool_body=~s/\x0a{2}$//; # Cut off 2 trailing LF's
		$_[2]=$spool_body; 
		$_[3]=1; 
		return %spool_header; }
	    %spool_header=(); 
	    $header_done=0; 
	    $first_body_line=0; 
	    $spool_body=""; 
	}
	# The following message parsing comes from revokelib.pl...they
        # should pretty much match to be consistant.
	if (($line=~/^$/) && (!$header_done)) # a line of whitespace
	{ $header_done=1; $first_body_line=1; }
	if ($header_done) 
	{   # Get the body and remove leading whitespace
	    if ($first_body_line)
	    { $first_body_line=0; }
	    else 
	    {$spool_body= $spool_body.$line; }
	} else {
            # Header field names may contain any ASCII chacter from 
            # 33->126 (decimal) except colon
	    # Note that this does not catch all headers...if the header
	    # is multiline, or stuff like To: Name <fubob@foo.com> won't work
	    if ($line=~/^([\x21-\x7E]*):\s*(.*)/i)
	    { 
		$spool_header{$1}=$2; 
	    }
	}	
    }
    
    # Check if last message is the "chosen one".
    if ($hash eq $spool_header{'X-Revoke-Hash'}) # Found it!
    {
	$spool_body=~s/\x0a{2}$//; # Cut off 2 trailing LF's
	$_[2]=$spool_body; 
	$_[3]=1; 
	return %spool_header; }
    $_[3]=0;
    ();
}

sub reply # ($message,$body,$separator,%header)
# Effects: Sends a message specifying whether the message was found or not.
#          It's sent to the Reply-To or From address (in that order of prec.)
{
    local($message,$body,$separator,%header)=@_;
    local($to,$temp,$path,$draft);
    if ($header{'Reply-To'}) 
    { $to=$header{'Reply-To'}; }
    else
    { $to=$header{'From'}; }
    
    chop($temp=`$mhparam_path mhtemp`);
    if ($temp=~/^$/) { $temp="temp"; }
    chop($path=`$mhparam_path path`);
    if ($path=~/^$/) { $path="Mail"; }
    $draft="$ENV{'HOME'}/$path/$temp";
    
    open(DRAFT,">$draft");
    print DRAFT "To: $to\nSubject: Status of message\n------\n\n";
    print DRAFT "$message\n\n";
    print DRAFT "Your request:\n\n";
    while (($key,$value)=each(%header))
    { print DRAFT ">$key: $value\n"; }
    print DRAFT "$separator\n";
    print DRAFT "$body\n";
    close (DRAFT);
    `$send_path $temp`;  # Send uses files wrt the user's $path directory
}

sub replace_hash # ($spool,$oldhash,$newhash)
# Requires: $spool is the path to the user's mail spool file and $oldhash 
#           appears at most once as a header.   We assume the likelihood
#           of two messages with the same encrypted hash is null.
#           /tmp must exist.
# Effects: Searches for "X-Revoke-Hash: $oldhash" and replaces
#          the oldhash with the new one.
# Note: This will bomb if a non-quoted X-Revoke-Hash header starts a line 
#       in the body of the message.
{
    local($spool,$oldhash,$newhash)=@_;
    local($temp)="/tmp/$ENV{'USER'}.spool.$$";
    open(TEMPSPOOL,">$temp");
    open(SPOOL,$spool);    
    #flock(SPOOL,$LOCK_EX);
    # For each occurance of the old hash (hopefully only once!), replace it.
    while(<SPOOL>)
    {
	# Don't forget that \s matches \n too!  It's a Brave New Line!
	s/^X-Revoke-Hash:\s*$oldhash\s*$/X-Revoke-Hash: $newhash\n/;
	print TEMPSPOOL;
    }
    open(SPOOL,">$spool");
    open(TEMPSPOOL,"$temp");
    while (<TEMPSPOOL>) { print SPOOL; }
    #flock(SPOOL,$LOCK_UN);
    close(TEMPSPOOL);
    `$rm_path $temp`;
    close(SPOOL);
}

sub remove_message # ($spool,$hash)
# Requires: $hash occurs at most once in spool file.  This hash is the *old*
#           hash.
# Effects: Removes the message that contains the X-Revoke-Hash header that
#          matches.
{
    local($spool,$hash)=@_;
    local(@message,$buffered_line,$revokeme);
    open(SPOOL,$spool);
    local($temp)="/tmp/$ENV{'USER'}.spool.$$"; # $$ is the PID
    open(TEMP,">$temp"); 
    $revokeme=0;
    while ($line=<SPOOL>)
    {
	if ($line=~/^From\x20/)
	{   print REVOKELOG "Found a from: >$line<\n";
            # Beginning of next message, flush old message
	    if ($revokeme)
	    {   # Exterminate!  Exterminate!
		@message=();
		$revokeme=0;
	    }
	    else
	    {   # The previous message, if any, wasn't the one to revoke.
		# If this is the first message, nothing will get printed
		# anyway.
		foreach $buffered_line (@message)
		{ print TEMP "$buffered_line"; }
		@message=();
	    }
	}
	if ($line=~/^X-Revoke-Hash:\s*$hash/i)
	{ $revokeme=1; } # This message is to be erradicated!
	push(@message,$line);
    }
    if (! $revokeme)
    {	
	foreach $buffered_line (@message)
	{ print TEMP "$buffered_line"; }
    }

    # Move the re-created spool to the actual spool file (paste over)
    open(TEMP,"$temp");
    open(SPOOL,">$spool");
    while (<TEMP>) { print SPOOL; }
    close(SPOOL);
    close(TEMP);
    `$rm_path $temp`;
}
1;





